import plotly.graph_objects as go

from plotly.offline import iplot
from pysdf import SDF
from warnings import warn

from echoviz.utils.colors import BIN_CMAPS, HEAT_CMAPS, SJET, SJET_R
from echoviz.utils.layouts import LAYOUT_3D, LAYOUT_4D, CONTOUR, frame_args
from echoviz.utils.misc import clean_fname



def animated_3d(vinputs, vlabels=None, vpreds=None, threshold=None,
                title="", show=True, filename=None):
    frames = []
    #TODO: Multithread this
    for i in range(len(vinputs)):
        data = [vinputs[i].make_mesh(opacity=0.3, showscale=False, colorscale="Greys", cmin=0, cmax=1)]
        if vlabels:
            for k in vlabels.keys():
                try:
                    data.append(vlabels[k][i].make_mesh(color=BIN_CMAPS[k][1]))
                except RuntimeError:
                    warn(f"{k.capitalize()} label for frame {i} is null, not plotting it.", RuntimeWarning)
        if vpreds:
            for k in vpreds.keys():
                try:
                    if threshold:
                        data.append(vpreds[k][i].float2bool(threshold)
                                                .make_mesh(color=BIN_CMAPS[k][1]))
                    else:
                        data.append(vpreds[k][i].make_mesh(colorscale=HEAT_CMAPS[k], cmin=0, cmax=1))
                except RuntimeError:
                    warn(f"{k.capitalize()} prediction for frame {i} is null, not plotting it.", RuntimeWarning)
        frames.append(go.Frame(data=data, name=str(i)))
    layout = LAYOUT_4D
    layout["sliders"][0]["steps"] = [{
        "args": [[str(i)], frame_args(0)],
        "label": str(i),
        "method": "animate",
        } for i in range(len(frames)) ]
    #FIXME: fix range to smooth animation
    fig = go.Figure(data=frames[0].data, frames=frames, layout=layout)
                    #range_x=[], range_y=[], range_z=[])
    fig.update_layout(title=title, **LAYOUT_3D)
    fig.update_traces(showscale=False)
    if show:
        iplot(fig)
    if filename:
        filename = clean_fname(filename)
        fig.write_html(filename) #FIXME: Reduce file size
    return fig



def _dist_4d(vlabels, vpreds, fdist, vinputs=None,
             title='', colorscale=SJET_R, show=True, filename=None):
    frames = []
    cmin, cmax = 10000, -10000
    nbf = len(list(vlabels.values())[0])
    #TODO: Multithread this
    for i in range(nbf):
        data = []
        if vinputs:
            data.append(vinputs[i].make_mesh(opacity=0.3, showscale=False, colorscale="Greys", cmin=0, cmax=1))
        for k in vlabels.keys():
            try:
                verts, faces, _ = vlabels[k][i].devoxelize()
            except RuntimeError:
                warn(f"{k.capitalize()} label for frame {i} is null, not plotting it.", RuntimeWarning)
            sdf = SDF(verts, faces)
            try:
                verts, faces, _ = vpreds[k][i].devoxelize()
            except RuntimeError:
                warn(f"{k.capitalize()} prediction for frame {i} is null, not plotting it.", RuntimeWarning)
            sdf_res = fdist(sdf(verts))
            cmin, cmax = min(min(sdf_res), cmin), max(max(sdf_res), cmax)
            data.append(vpreds[k][i].make_mesh(verts=verts, faces=faces, values=sdf_res, name=k,
                                               opacity=0.7, colorscale=colorscale, contour=CONTOUR))
        frames.append(go.Frame(data=data, name=str(i)))
    layout = LAYOUT_4D
    layout["sliders"][0]["steps"] = [{
        "args": [[str(i)], frame_args(0)],
        "label": str(i),
        "method": "animate",
        } for i in range(len(frames)) ]
    #FIXME: fix range to smooth animation
    fig = go.Figure(data=frames[0].data, frames=frames, layout=layout)
                    #range_x=[], range_y=[], range_z=[])
    fig.update_layout(title=title, **LAYOUT_3D)
    fig.update_traces({"cmin": cmin, "cmax": cmax, "showscale": True}, lambda t: t.name != "input")
    if show:
        iplot(fig)
    if filename:
        filename = clean_fname(filename)
        fig.write_html(filename) #FIXME: Reduce file size
    return fig


def sdf_animated_3d(vlabels, vpreds, vinputs=None, title='', show=True, filename=None):
    return _dist_4d(vlabels, vpreds, lambda x: x, vinputs=vinputs, title=title,
                    show=show, filename=filename)

def asd_animated_3d(vlabels, vpreds, vinputs=None, title='', show=True, filename=None):
    return _dist_4d(vlabels, vpreds, lambda x: abs(x), vinputs=vinputs, title=title,
                    colorscale=SJET_R, show=show, filename=filename)



LAYOUT_2D = {"title_y": 0.98,
             "autosize": False,
             "height": 350,
             "margin": {'l': 10, 'r': 10, 't': 30, 'b': 10}
             }

NAKED_AXIS = {"showgrid": False, "showticklabels": False}


CAMERA = {"eye": {'x': -0.5, 'y': -1.15, 'z': 0.7},
          "up": {'x': 0, 'y': 0, 'z': 0.5}}

LAYOUT_3D = {"scene_camera": CAMERA,
             "plot_bgcolor": "rgb(64, 64, 64)",
             "title_y": 0.98,
             "autosize": False,
             "width": 600,
             "height": 500,
             "margin": {'l': 10, 't': 30, 'b': 10}
             }


CONTOUR = {"color": "lightgrey", "show": True, "width": 10}


def frame_args(duration):
    return {"frame": {"duration": duration, "redraw": True},
            "mode": "immediate",
            "fromcurrent": True,
            "transition": {"duration": duration, "easing": "linear"}}

LAYOUT_4D = {
        "sliders": [{
            "len": 0.8,
            'x': 0.1, 'y': 0
            }],
        "updatemenus": [{
            "buttons": [{
                "args": [None, frame_args(200)],
                "label": "&#9654;", # Play symbol
                "method": "animate"
                },
                {
                "args": [[None], frame_args(0)],
                "label": "&#9724;", # Pause symbol
            "method": "animate"}],
            "direction": "left",
            "type": "buttons",
            'x': 0, 'y': 0
            }],
        }



"""
3D plot of mitral valve voxels from voxel grids (see `utils/voxel_grid.py`)
"""

import plotly.express as px

from io import BytesIO
from PIL import Image

from echoviz.threed.distance import sdf_interactive_3d, asd_interactive_3d
from echoviz.threed.interactive import interactive_3d
from echoviz.utils.misc import clean_fname



def _screenshot_3d(fig, show, filename):
    extension = str(filename).split('.')[-1] if filename else "png"
    img_str = fig.to_image(format=extension)
    if show:
        img = Image.open(BytesIO(img_str))
        fig = px.imshow(img)
        fig.update_layout(autosize=False, width=600, height=500,
                          margin=dict(l=0, t=0, r=0, b=0))
        fig.update_xaxes(showticklabels=False)
        fig.update_yaxes(showticklabels=False)
        fig.show()
    if filename:
        filename = clean_fname(filename, f".{extension}")
        with open(filename, "wb") as fd:
            fd.write(img_str)


def static_3d(vinput, vlabels=None, vpreds=None, threshold=None,
              title="", show=True, filename=None):
    fig = interactive_3d(vinput, vlabels, vpreds, threshold, title, show=False)
    _screenshot_3d(fig, show, filename)


def sdf_static_3d(vlabels, vpreds, vinputs=None, title="", show=True, filename=None):
    fig = sdf_interactive_3d(vlabels, vpreds, vinputs, title, show=False)
    _screenshot_3d(fig, show, filename)

def asd_static_3d(vlabels, vpreds, vinputs=None, title="", show=True, filename=None):
    fig = asd_interactive_3d(vlabels, vpreds, vinputs, title, show=False)
    _screenshot_3d(fig, show, filename)



"""
3D plot of mitral valve voxels from voxel grids (see `utils/voxel_grid.py`)
Functions are adaptated from: https://www.raddq.com/dicom-processing-segmentation-visualization-in-python/
"""

import numpy as np
import plotly.graph_objects as go

from plotly.offline import iplot
from warnings import warn

from echoviz.utils import BIN_CMAPS, HEAT_CMAPS
from echoviz.utils.layouts import LAYOUT_3D
from echoviz.utils.misc import clean_fname



def interactive_3d(vinput, vlabels=None, vpreds=None, threshold=None,
                   title="", show=True, filename=None):
    """
    """
    #FIXME? Add classes legend
    #FIXME: Make vinput optional
    if vlabels and vpreds:
        warn("Plotting label and annotation in 3D make for a busy plot", UserWarning)
    inp_mesh = vinput.make_mesh(showscale=False, opacity=0.3, colorscale="Greys", cmin=0, cmax=1)
    fig = go.Figure(data=inp_mesh, layout=dict(title=title, **LAYOUT_3D))
    if vlabels:
        lab_meshes = []
        for k in vlabels.keys():
            try:
                lab_meshes.append(vlabels[k].make_mesh(color=BIN_CMAPS[k][1]))
            except RuntimeError:
                warn(f"{k.capitalize()} label is null, not plotting it.", RuntimeWarning)
        fig.add_traces(lab_meshes)
    if vpreds:
        pred_meshes = []
        for k in vpreds.keys():
            try:
                if threshold:
                    pred_meshes.append(vpreds[k].float2bool(threshold)
                                                .make_mesh(color=BIN_CMAPS[k][1]))
                else:
                    pred_meshes.append(vpreds[k].make_mesh(colorscale=HEAT_CMAPS[k], cmin=0, cmax=1))
            except RuntimeError:
                warn(f"{k.capitalize()} prediction is null, not plotting it.", RuntimeWarning)
        fig.add_traces(pred_meshes)
        fig.update_traces(showscale=False)
    if show:
        iplot(fig)
    if filename:
        filename = clean_fname(filename)
        #FIXME: Diminish file size by removing js modules
        fig.write_html(filename)
    return fig



import h5py

from datetime import timedelta
from pathlib import Path
import subprocess
from time import sleep, time

from echoviz import VoxelGrid




POUTS = {ext: Path(f"test-output.{ext}").resolve() for ext in ["png", "gif", "html"]}
YES = ['y', "yes", '']



def time_plot(func, inputs, labels, *args, **kwargs):
    start = time()
    func(inputs, labels, *args, **kwargs)
    end = time()
    print(f"    `{func.__name__}` took {timedelta(seconds=(end - start))} to run.")

def show_res(dtype, path):
    sentence = ("      Do you want to see the result from previous test as"
                f" {dtype}? (Y/n) ")
    if input(sentence).lower() in YES:
        #FIXME: Don't assume firefox
        subprocess.call(["firefox", path.as_uri()])


def the_end(msg):
    print(msg)
    sleep(3) # Wait so last file can be seen before being removed
    for p in POUTS.values(): #Remove created files
        p.unlink(missing_ok=True)



def _get_grid(array, origin, directions, spacing):
    # You want to copy in case you change the scale because it's the same for several grids
    return VoxelGrid(array, origin.copy(), directions.copy(), spacing.copy())

def load_file(fname):
    hdf = h5py.File(fname, 'r')
    origin = hdf["VolumeGeometry"]["origin"][()]
    directions = hdf["VolumeGeometry"]["directions"][()]
    spacing  = hdf["VolumeGeometry"]["resolution"][()]
    vinp = []
    vtg, vpr = {"all": [], "anterior": [], "posterior": []}, {"all": [], "anterior": [], "posterior": []}
    multi = False
    for i in range(1, len(hdf["Input"]) + 1):
        vinp.append(_get_grid(hdf["Input"][f"vol{i:02d}"][()], origin, directions, spacing))
        try:
            vtg["all"].append(_get_grid(hdf["Target"][f"vol{i:02d}"][()],origin, directions, spacing))
            vpr["all"].append(_get_grid(hdf["Prediction"][f"vol{i:02d}"][()], origin, directions, spacing))
        except KeyError:
            for k in ["anterior", "posterior"]:
                vtg[k].append(_get_grid(hdf["Target"][f"{k}-{i:02d}"][()], origin, directions, spacing))
                vpr[k].append(_get_grid(hdf["Prediction"][f"{k}-{i:02d}"][()], origin, directions, spacing))
            multi = True
    hdf.close()
    if multi: # Clean created directories
        vtg.pop("all"), vpr.pop("all")
    else:
        vtg.pop("anterior"), vtg.pop("posterior")
        vpr.pop("anterior"), vpr.pop("posterior")
    if len(vinp) == 1: # Flatten if only one frame
        vinp = vinp[0]
        for k in vtg.keys():
            vtg[k] = vtg[k][0]
            vpr[k] = vpr[k][0]
    return vinp, vtg, vpr




from io import BytesIO
from plotly.offline import iplot
from PIL import Image

from echoviz.twod.slicing import plot_slice



def _get_frame(data, idx):
    if data:
        return {k: data[k][idx] for k in data.keys()}
    return None


def sliced_sequence(vinputs, vlabels, index, axis=1, vpreds=None, threshold=None,
                   plot_input=True, title='', filename="sliced_sequence.gif"):
    imgs = []
    for f, vinp in enumerate(vinputs): #FIXME: Multithread to optimize time
        vlab, vpred = _get_frame(vlabels, f), _get_frame(vpreds, f)
        fig = plot_slice(vinp, vlab, index, axis, vpred, threshold,
                         plot_input, title, show=False, filename=None)
        img_str = fig.to_image(format="png")
        imgs.append(Image.open(BytesIO(img_str)))
    with open(filename, "wb") as fd:
        imgs[0].save(fd, save_all=True, append_images=imgs[1:],
                     optimize=True, duration=100, loop=0)


def sliced_volume(vinput, vlabels, axis=1, vpreds=None, threshold=None,
                  stride=10, plot_input=True, title='', filename="sliced_volume.gif"):
    imgs = []
    max_index = vinput.shape[axis]
    #FIXME: Multithread to optimize time
    for index in range(0, max_index, stride):
        fig = plot_slice(vinput, vlabels, index, axis, vpreds, threshold,
                         plot_input, title, show=False, filename=None)
        img_str = fig.to_image(format="png")
        imgs.append(Image.open(BytesIO(img_str)))
    with open(filename, "wb") as fd:
        imgs[0].save(fd, save_all=True, append_images=imgs[1:],
                     optimize=True, duration=150, loop=0)



import plotly.graph_objects as go

from plotly.offline import iplot
from pysdf import SDF
from warnings import warn

from echoviz.utils.colors import SJET, SJET_R
from echoviz.utils.layouts import LAYOUT_3D, CONTOUR
from echoviz.utils.misc import clean_fname



def _dist_interactive_3d(vlabels, vpreds, fdist, vinputs=None,
                         title='', colorscale=SJET_R, show=True, filename=None):
    """
    """
    # Create fancy figure
    fig = go.Figure(layout=dict(title=title, **LAYOUT_3D))
    if vinputs:
        fig.add_trace(vinputs.make_mesh(showscale=False, name="input", opacity=0.3,
                                        colorscale="Greys", cmin=0, cmax=1))
    cmin, cmax = 10000, -10000
    for k in vlabels.keys():
        try:
            verts, faces, _ = vlabels[k].devoxelize()
        except RuntimeError:
            warn(f"{k.capitalize()} label is null, not plotting it.", RuntimeWarning)
        sdf = SDF(verts, faces)
        try:
            verts, faces, _ = vpreds[k].devoxelize()
        except RuntimeError:
            warn(f"{k.capitalize()} prediction is null, not plotting it.", RuntimeWarning)
        sdf_res = fdist(sdf(verts))
        cmin, cmax = min(min(sdf_res), cmin), max(max(sdf_res), cmax)
        mesh = vpreds[k].make_mesh(verts=verts, faces=faces, values=sdf_res, name=k,
                                   opacity=0.7, colorscale=colorscale, contour=CONTOUR)
        fig.add_trace(mesh)
    # Homogenize SDF colorscale
    fig.update_traces({"cmin": cmin, "cmax": cmax}, lambda t: t.name != "input")
    if show:
        iplot(fig)
    if filename:
        filename = clean_fname(filename)
        #FIXME: Diminish file size by removing js modules
        fig.write_html(filename)
    return fig


def sdf_interactive_3d(vlabels, vpreds, vinputs=None, title='', show=True, filename=None):
    return _dist_interactive_3d(vlabels, vpreds, lambda x: x, vinputs=vinputs,
                                title=title, show=show, filename=filename)

def asd_interactive_3d(vlabels, vpreds, vinputs=None, title='', show=True, filename=None):
    return _dist_interactive_3d(vlabels, vpreds, lambda x: abs(x), vinputs=vinputs,
                                title=title, colorscale=SJET, show=show, filename=filename)



import click as cli
import echoviz as ecv
import random as rd
import signal
import sys

from pathlib import Path

from test_utils import *




signal.signal(signal.SIGINT, lambda *args: (the_end("\nOk byye."), sys.exit(0)))



@cli.command(context_settings={"help_option_names": ["-h", "--help"], "show_default": True})
@cli.argument("fname", type=cli.Path(exists=True, resolve_path=True, path_type=Path, file_okay=True))
def testy_testa(fname):
    vinp, vtg, vpr = load_file(fname)
    if not isinstance(vinp, list):
        print("The given file is not a sequence, so I'll only test plot that runs on one frame.")
        frame = 1
        bvinp, bvtg, bvpr = vinp, vtg, vpr
    else:
        frame = rd.randint(0, len(vinp) - 1)
        bvinp = vinp[frame]
        bvtg, bvpr = { k: vtg[k][frame] for k in vtg.keys() }, { k: vpr[k][frame] for k in vpr.keys() }
    print(f"We're studying file {fname.name} at frame {frame}.")
    print("Let's start with some 2D static plots.")
    time_plot(ecv.plot_slice, bvinp, bvtg, 64, axis=1, plot_input=False, title="Labels on input")
    time_plot(ecv.plot_slice, bvinp, bvtg, 64, axis=0, vpreds=bvpr, threshold=0.5,
              filename=POUTS["png"], title="Labels on input & predictions")
    show_res("image", POUTS["png"])
    print("Now for some 3D static plots.")
    time_plot(ecv.static_3d, bvinp, bvtg, vpreds=bvpr, threshold=0.5,
              filename=POUTS["png"], title="Input, labels & predictions")
    show_res("image", POUTS["png"])
    # SDF (m)
    time_plot(ecv.sdf_static_3d, bvtg, bvpr, vinputs=bvinp, title="SDF (m)")
    # ASD (m)
    bvinp.set_scale('m')
    for k in bvtg.keys():
        bvtg[k].set_scale('m'), bvpr[k].set_scale('m')
    time_plot(ecv.asd_static_3d,bvtg, bvpr, show=False, filename=POUTS["png"], title="ASD (m)")
    show_res("image", POUTS["png"])
    print("Let's take it up a notch and have an interactive 3D plot!")
    time_plot(ecv.interactive_3d, bvinp, bvtg, show=False, filename=POUTS["html"],
              title="Input & labels")
    show_res("web file", POUTS["html"])
    #SDF (mm)
    bvinp.set_scale("mm")
    for k in bvtg.keys():
        bvtg[k].set_scale("mm"), bvpr[k].set_scale("mm")
    time_plot(ecv.sdf_interactive_3d, bvtg, bvpr, vinputs=bvinp, title="SDF (mm)")
    # ASD (m)
    bvinp.set_scale('m')
    for k in bvtg.keys():
        bvtg[k].set_scale('m'), bvpr[k].set_scale('m')
    time_plot(ecv.asd_interactive_3d, bvtg, bvpr, title="ASD (m)", show=False, filename=POUTS["html"])
    show_res("web file", POUTS["html"])
    print("Back to 2D! But with a twist... It's animated o/")
    time_plot(ecv.sliced_volume, bvinp, bvtg, axis=2, vpreds=bvpr, stride=5,
              plot_input=False, filename=POUTS["gif"], title="Labels on input & predictions")
    show_res("gif", POUTS["gif"])
    if not isinstance(vinp, list):
        the_end("You didn't give a sequence, so I can't further unfortunately. Hope you still had fun!")
        return
    time_plot(ecv.sliced_sequence, vinp, vtg, 64, axis=0, filename=POUTS["gif"], title="Input & labels")
    show_res("gif", POUTS["gif"])
    print("And for the grand finale... Interactive 3D animated plot!!")
    time_plot(ecv.animated_3d, vinp, None, vpreds=vpr,
              filename=POUTS["html"], title="Input & predictions", show=False)
    show_res("web file", POUTS["html"])
    #SDF (m)
    time_plot(ecv.sdf_animated_3d, vtg, vpr, vinputs=vinp, title="SDF (m)", show=False, filename=POUTS["html"])
    show_res("web file", POUTS["html"])
    # ASD (mm)
    for i in range(len(vinp)):
        vinp[i].set_scale("mm")
    for k in vtg.keys():
        for i in range(len(vtg[k])):
            vtg[k][i].set_scale("mm"), vpr[k][i].set_scale("mm")
    time_plot(ecv.asd_animated_3d, vtg, vpr, title="ASD (mm)", show=True)
    the_end("That's all folks!")


if __name__ == "__main__":
    testy_testa()



from echoviz.threed.animated import animated_3d, sdf_animated_3d, asd_animated_3d
from echoviz.threed.distance import sdf_interactive_3d, asd_interactive_3d
from echoviz.threed.interactive import interactive_3d
from echoviz.threed.static import static_3d, sdf_static_3d, asd_static_3d



from setuptools import setup



with open("echoviz/__init__.py", 'r') as fd:
    for line in fd:
        if line.startswith("__version__"):
            VERSION = line.split('=')[-1].strip()


setup(name="echoviz",
      version=VERSION,
      url="https://github.com/mailys-hau/echoviz",
      author="Maïlys HAU",
      packages=["echoviz", "echoviz.threed", "echoviz.twod", "echoviz.utils"],
      install_requires=["kaleido", "numpy", "pandas", "pillow", "plotly", "pysdf",
                        "scikit-image", "ipython"],
     )



from echoviz.threed import (animated_3d, sdf_animated_3d, asd_animated_3d,
                            interactive_3d, sdf_interactive_3d, asd_interactive_3d,
                            static_3d, sdf_static_3d, asd_static_3d)
from echoviz.twod import plot_slice, sliced_sequence, sliced_volume
from echoviz.utils import VoxelGrid



__author__ = "Maïlys Hau"
__version__ = "0.2.1"
